/*
 * Copyright (c) 2005 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <linux/joystick.h>

#include "joystick.h"

#define NAME_LENGTH 256
/*----------------------------------------------------------------------*/
/**
 ** Init joystick 
 **/
JOYSTICK_STR *
joystickInit(char *dev)
{
	int joy;
	int version;
	unsigned char axes, buttons;
	JOYSTICK_STR *j;
	char name[NAME_LENGTH];

	j = (JOYSTICK_STR *)calloc(1, sizeof(JOYSTICK_STR));
	if (j == NULL)
		return NULL;
	
	joy = open(dev, O_RDONLY);
	if (joy < 0) {
		perror(dev);
		return NULL;
	}
	ioctl(joy, JSIOCGVERSION, &version);
	ioctl(joy, JSIOCGAXES, &axes);
	ioctl(joy, JSIOCGBUTTONS, &buttons);
	ioctl(joy, JSIOCGNAME(NAME_LENGTH), name);
	
	printf("Joystick (%s) has %d axes and %d buttons.\n"
	       "\tDriver version is %d.%d.%d.\n",
	       name, axes, buttons, version >> 16, 
	       (version >> 8) & 0xff, version & 0xff);
	if (fcntl(joy, F_SETFL, O_NONBLOCK) < 0) {
		fprintf(stderr, "error setting non blocking mode on %s: %s\n",
			dev, strerror(errno));
		close(joy);
		free(j);
		return NULL;
	}
	j->fd = joy;
	return j;
}

/*----------------------------------------------------------------------*/

void
joystickEnd(JOYSTICK_STR *j)
{
	close(j->fd);
	free(j);
}

/*----------------------------------------------------------------------*/

int
joystickPoll(JOYSTICK_STR *j)
{
	struct js_event js;
	ssize_t n;
	int ret = 0;

	do {
		n = read(j->fd, &js, sizeof(struct js_event));
		if (n == -1) {
			if (errno == EAGAIN)
				return ret;
			else {
				printf("joystickPoll: %s\n", strerror(errno));
				return -1;
			}
		} else if (n == 0) {
			return ret;
		} else if (n != sizeof(struct js_event)) {
			/* short read */
			return ret;
		}
		/* n == sizeof(struct js_event) */
		switch (js.type & ~JS_EVENT_INIT) {
		case JS_EVENT_BUTTON:
			j->buttons[js.number] = js.value;
			break;
		case JS_EVENT_AXIS:
			j->axes[js.number] = js.value;
			break;
		} /* switch */
		ret++;
	} while (1);
	return ret;
}

/*----------------------------------------------------------------------*/
		
		

	      
