/*
 * Copyright (c) 2009-2013 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/**
 ** rmp440MotionTaskCodels.c
 **
 ** Codels called by execution task rmp440MotionTask
 **
 ** Author: Matthieu Herrb
 ** Date: April 2009
 **
 **/
#include <sys/types.h>
#include <errno.h>
#include <math.h>
#include <pthread.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>

#include <portLib.h>
#include <h2mathLib.h>
#include "h2timeLib.h"

#include <genPos/genPosStruct.h>
#include <contlaw/contLaw.h>
#include <rmp440/rmp440.h>

#include "rmp440Const.h"
#include "server/rmp440Header.h"

#ifdef WITH_FELIB
#include <fe/ftdi-emergency.h>
#endif

#include "odoProba.h"
#include "joystick.h"

#ifdef GYRO
#include <gyroLib/gyro.h>
#endif

static POSTER_ID pomPosterId = NULL;
static JOYSTICK_STR *joy = NULL;
static struct rmp440_feedback *data = NULL;
static struct rmp440_io *rmp = NULL;
#ifdef WITH_FELIB
static FE_STR *fe = NULL;
#endif
#ifdef GYRO
static struct GYRO_DATA *gyroId = NULL;
#endif

extern void *rmp440ReadTask(void *);
extern STATUS rmp440Odo3d(int *report, double period);

static STATUS rmp440DataUpdate(RMP440_STATUS_STR *);
static STATUS rmp440VelocitySet(double, double, int *);
static STATUS initPomPosters(void);
static STATUS updatePomPosters(void);
static STATUS rmp440VelocityGet(void);
static void getJoystickSpeeds(double *v, double *w, double *avmax, double *awmax);


/*------------------------------------------------------------------------
 *
 * initOdoAndAsserv  --  Initialization codel (fIDS, ...)
 *
 * Description:
 *
 * Returns:    OK or ERROR
 */

STATUS
initOdoAndAsserv(int *report)
{
	GENPOS_CART_STATE *configId;
	GENPOS_CART_REF *refId;
	GENPOS_CART_CMD_PARAM *cmdId;
	RMP440_KINEMATICS_STR *kinematics;
	RMP440_GYRO *gyro;
	RMP440_GYRO_ASSERV *gyroAsserv;

	/* Kinematics */
	kinematics = &rmp440DataStrId->kinematics;
	kinematics->leftWheelRadius = RMP_X2_TIRE_DIAMETER/2.0;
	kinematics->rightWheelRadius = RMP_X2_TIRE_DIAMETER/2.0;
	kinematics->axisWidth = RMP_DEFAULT_WHEEL_TRACK_WIDTH;

	/* configuration */
	configId = &rmp440DataStrId->robot;
	configId->xRob = 0.;
	configId->yRob = 0.;
	configId->theta = 0.;
	configId->xRef = 0;
	configId->yRef = 0;
	configId->v = 0;
	configId->w = 0;

	/* Ref Data */
	refId = &rmp440DataStrId->ref;
	refId->dataType = GENPOS_POS_AND_SPEED_DATA;
	refId->backFlag = GENPOS_FORWARD_MOTION;
	refId->x = configId->xRef;
	refId->y = configId->yRef;
	refId->theta = configId->theta;
	refId->v = 0.;
	refId->w = 0.;

	refId->vmax = RMP_DEFAULT_MAXIMUM_VELOCITY;
	refId->wmax = RMP_DEFAULT_MAXIMUM_YAW_RATE;
	refId->linAccelMax = RMP_DEFAULT_MAXIMUM_ACCEL;
	refId->angAccelMax = RMP_DEFAULT_MAXIMUM_YAW_ACCEL;

	/* Control parameters */
	cmdId = &rmp440DataStrId->cmd;
	cmdId->distPoint = RMP440_DIST_TO_CONTROLLED_POINT;
	cmdId->KpLongit = RMP440_KP_LONGIT;
	cmdId->KiLongit = RMP440_KI_LONGIT;
	cmdId->KpTransv = RMP440_KP_TRANSV;
	cmdId->KiTransv = RMP440_KI_TRANSV;

	cmdId->KpAng = RMP440_KP_ANG;
	cmdId->KiAng = RMP440_KI_ANG;

	cmdId->longitMaxGap = RMP440_LONGIT_MAX_GAP;
	cmdId->transvMaxGap = RMP440_TRANSV_MAX_GAP;
	cmdId->angMaxGap    = RMP440_ANG_MAX_GAP;

	cmdId->longitErrorMax = RMP440_LONGIT_ERROR_MAX;
	cmdId->transvErrorMax = RMP440_TRANSV_ERROR_MAX;
	cmdId->angErrorMax = RMP440_ANG_ERROR_MAX;

	cmdId->longitMaxErrSum = RMP440_LONGIT_MAX_ERR_SUM;
	cmdId->transvMaxErrSum = RMP440_TRANSV_MAX_ERR_SUM;
	cmdId->angMaxErrSum = RMP440_ANG_MAX_ERR_SUM;

	rmp440DataStrId->vCommand = rmp440DataStrId->wCommand = 0.0;

	/* gyro */
	gyro = &rmp440DataStrId->gyro;
	gyro->params.mode = RMP440_GYRO_ON;
#ifdef GYRO
	gyro->gyroOn = FALSE;
#endif
	gyro->gyroToRobotOffset = 0.0;
	gyro->gyroTheta = 0.0;

	/* gyro asserv */
	gyroAsserv = &rmp440DataStrId->gyroAsserv;
	gyroAsserv->enabled = 0;
	gyroAsserv->first = 1;
	gyroAsserv->straight = 0;

	/* max accel */
	rmp440DataStrId->maxAccel.prev_vel_command = 0.;
	rmp440DataStrId->maxAccel.prev_vel_command_t = -1.;

	rmp440DataStrId->wheelSizeCorrection = 1.05;

	/* Initialize Pom */
	initPomPosters();

	return OK;
}

/*----------------------------------------------------------------------*/
/*
 * rmp440DataUpdate - copy data from 2 motors block to the IDS
 */
static STATUS
rmp440DataUpdate(RMP440_STATUS_STR *status)
{
	static float prevFrame = 0;
	static int noData = 0;

	if (data->frame_count == prevFrame)
		noData++;
	else
		prevFrame = data->frame_count;

	/* If not data for several periods - set motors off */
	if (noData >  10) {
		printf("No data ?\n");
		// status->rs_mode = RMP440_MOTORS_OFF;
	}
	if (noData)
		return OK;

	if (data->operational_state != 4)
		status->rs_mode = RMP440_MOTORS_OFF;
	status->operational_state = data->operational_state;

	status->propulsion_battery_level = data->min_propulsion_batt_soc;
	status->aux_battery_level = data->aux_batt_soc;
	status->pitch = data->pse_pitch;
	status->roll = data->pse_roll;
	status->yaw_rate = data->pse_yaw_rate;
	status->v = data->linear_vel;
	status->w = data->differential_wheel_vel /
	    data->fram_wheel_track_width;

#ifdef WITH_FELIB
	/* Check emergency stop */
	fe_get_status(fe);
	/* Check if power cord (dis)connected */
	if ((fe_pins(fe) & FE_POWER_CORD) != 0) {
		if (status->rs_mode != RMP440_POWER_CORD) {
			printf("Power Cord connected\n");
			status->rs_mode = RMP440_POWER_CORD;
		}
	} else if (status->rs_mode == RMP440_POWER_CORD) {
		printf("Power Cord disconnected\n");
		status->rs_mode = RMP440_IDLE;
	}
	/* ignore pause if motors are off */
	if ((fe_pins(fe) & FE_PAUSE) != 0 &&
	    status->rs_mode != RMP440_EMERGENCY &&
	    status->rs_mode != RMP440_MOTORS_OFF) {
		printf("Emergency pause!\n");
		if (status->rs_mode != RMP440_MOTORS_OFF)
			status->rs_mode = RMP440_EMERGENCY;
	}
#endif
	return OK;
}



/*------------------------------------------------------------------------
 *
 * odo  --  1st codel of permanent activity (called BEFORE other activities)
 *
 * Description: Reads all robots state and computes odometry
 *
 * Reports:      OK
 *              S_rmp440_EMERGENCY_STOP
 *
 * Returns:    OK or ERROR
 */

STATUS
odo(int *report)
{
	GENPOS_CART_REF *refId = &SDI_F->ref;
	GENPOS_CART_STATE *robot = &SDI_F->robot;
	RMP440_STATUS_STR *status = &SDI_F->status;
	RMP440_GYRO *gyro = &SDI_F->gyro;
	double direction;
#ifdef GYRO
	uint32_t date;
#endif
#if DEBUG>=1
	static int count = 0;
#endif

	if (rmp == NULL)
		return OK;		/* not initialized yet */

	rmp440ReceiveAndDecode(rmp, data);
	rmp440DataUpdate(status);

	/* Read config */
	if (rmp440VelocityGet() != OK)
		return ERROR;
	if (refId->backFlag == GENPOS_FORWARD_MOTION)
		direction = 1;
	else if (refId->backFlag == GENPOS_BACKWARD_MOTION)
		direction = -1;
	else
		direction = 0;
	robot->xRef = robot->xRob
		+ direction*RMP440_DIST_TO_CONTROLLED_POINT*cos(robot->theta);
	robot->yRef = robot->yRob
		+ direction*RMP440_DIST_TO_CONTROLLED_POINT*sin(robot->theta);

	odoProba(&SDI_F->robot, &SDI_F->var,
	    SDI_F->kinematics.axisWidth, SDI_F->varParams.coeffLinAng,
		   EXEC_TASK_PERIOD(RMP440_MOTIONTASK_NUM));

	/* Gyro */
#ifdef GYRO
	if (gyroId != NULL && gyro->currentMode != RMP440_GYRO_OFF) {
		if (!(gyroRead(gyroId, &gyro->gyroTheta,
			    &gyro->gyroOmega, &date))) {
			gyro->currentMode = RMP440_GYRO_OFF;
			gyroEnd(gyroId);
			gyroId = NULL;
		}
	}
#endif
	if (gyro->currentMode == RMP440_GYRO_OFF ||
	    (gyro->currentMode == RMP440_GYRO_ON_IF_MOTION
		&& fabs(robot->w) < 0.017
		&& fabs(robot->v) < 0.01)) {
		/* -- odo is reference */

		/* reset gyro offset to match odo */
		gyro->gyroToRobotOffset = angleLimit(robot->theta
		    - gyro->gyroTheta);
		/* switch back to odo */
		if (gyro->gyroOn) {
#if DEBUG>=1
			printf ("-> odo (%5.2f d) v %5.2f w%5.3f\n",
			    RAD_TO_DEG(robot->theta), robot->v, robot->w);
#endif
			gyro->gyroOn = FALSE;
		}
	} else {
		/* -- gyro is reference */
		robot->theta = angleLimit(gyro->gyroTheta +
		    gyro->gyroToRobotOffset);
		/* switch back to gyro */
		if (!gyro->gyroOn) {
#if DEBUG>=1
			printf ("-> gyro (%5.2f d) v %5.2f w%5.3f\n",
			    RAD_TO_DEG(robot->theta), robot->v, robot->w);
#endif
			gyro->gyroOn = TRUE;
		}
	}

	/*debug */
#if DEBUG>=1
	if (count==10) {
		printf ("MODE %s. gyro %5.3f + offset %5.3f = %5.3f =? %5.3f\n",
		    gyro->gyroOn ? "GYRO" : "ODO",
		    RAD_TO_DEG(gyro->gyroTheta),
		    RAD_TO_DEG(gyro->gyroToRobotOffset),
		    RAD_TO_DEG(angleLimit(gyro->gyroTheta +
			    gyro->gyroToRobotOffset)),
		    RAD_TO_DEG(robot->theta));
		count=0;
	}
	count++;
#endif
	if (SDI_F->odometryMode == RMP440_ODO_3D) {
	  rmp440Odo3d (report, EXEC_TASK_PERIOD(RMP440_MOTIONTASK_NUM));


	}
	return OK;
}

/*----------------------------------------------------------------------*/

/**
 ** Actual tracking code
 **/
static void
track(int *report, double *vRef, double *wRef)
{
	GENPOS_CART_REF *refId = &SDI_F->ref;
	GENPOS_CART_STATE *robot = &SDI_F->robot;
	GENPOS_TRACK *trackId = &SDI_F->trackStr;
	static double prevV, prevW;
	static BOOL prevDrifted = FALSE;
	BOOL drifted = FALSE;

	switch (trackId->trackMode) {
	case GENPOS_TRACK_POS:
		if (refId->dataType != GENPOS_POS_DATA) {
			fprintf(stderr, "wrong data type for TRACK_POS\n");
			*report = S_rmp440_BAD_REF;
			return;
		}
		drifted = contLawCartPositionControl(&rmp440DataStrId->cmd,
		    refId, robot, EXEC_TASK_PERIOD(RMP440_MOTIONTASK_NUM),
		    prevV, prevW, vRef, wRef);
		break;

	case GENPOS_TRACK_CONFIG:
		if (refId->dataType != GENPOS_POS_AND_SPEED_DATA) {
			fprintf(stderr, "wrong data type for TRACK_CONFIG\n");
			*report = S_rmp440_BAD_REF;
			return;
		}
		drifted = contLawCartConfigControl(&rmp440DataStrId->cmd,
		    refId, robot, EXEC_TASK_PERIOD(RMP440_MOTIONTASK_NUM),
		    prevV, prevW, vRef, wRef);
		break;

	case GENPOS_TRACK_SPEED:
		if (refId->dataType != GENPOS_SPEED_DATA) {
			fprintf(stderr, "wrong data type for TRACK_SPEED %d\n",
			    refId->dataType);
			*report = S_rmp440_BAD_REF;
			return;
		}
		*vRef = (refId->backFlag == GENPOS_FORWARD_MOTION ? refId->v :
		    -refId->v);
		*wRef = refId->w;
		break;
	default:
		*report = S_rmp440_BAD_REF;
		break;
	}
	if (drifted && !prevDrifted) {
		printf("odoAndAsserv: important drift!\n");
		printf("ref: %.2lf %.2lf %.2lf\n",
		       refId->x, refId->y, refId->theta);
		printf("pos: %.2lf %.2lf %.2lf\n",
		       robot->xRef, robot->yRef, robot->theta);
	}
	if (!drifted && prevDrifted)
		printf("odoAndAsserv: drift cancelled\n");
	prevDrifted = drifted;

	prevV = *vRef;
	prevW = *wRef;
}

/*----------------------------------------------------------------------*/

void
bound_accels(double t, double *vel_reference, double *ang_reference)
{
	RMP440_MAX_ACCEL *acc;
	acc = &rmp440DataStrId->maxAccel;

	// don't use real time but rather module's period, so that if some periods
	// are lost it doesn't accelerate too much afterwards

	const double epsilon = 1e-3;
	int keep_radius = (fabs(*ang_reference) > epsilon &&
	    fabs(*vel_reference) > epsilon);
	double radius;

	if (keep_radius)
		radius = *vel_reference / *ang_reference;

	// linear velocity
	double sign_accel = *vel_reference > acc->prev_vel_command ?
	    +1.0 : -1.0;
	if (acc->prev_vel_command * *vel_reference < 0)
		*vel_reference = 0.0;

	double max_accel = fabs(acc->prev_vel_command) < fabs(*vel_reference) ?
	    RMP_DEFAULT_MAXIMUM_ACCEL : RMP_DEFAULT_MAXIMUM_DECEL;
	double max_vel = acc->prev_vel_command +
	    sign_accel * max_accel * RMP440_SEC_PERIOD;
	*vel_reference = sign_accel *
	    fmin(sign_accel * *vel_reference, sign_accel * max_vel);
	acc->prev_vel_command = *vel_reference;

	// angular velocity to preserve radius
	if (keep_radius) *ang_reference = *vel_reference / radius;
}

/*----------------------------------------------------------------------*/

void
control_yaw(int *report, double t, double vel_reference, double yawr_reference,
	    double yawr_measure, double yaw_measure, double *yawr_command)
{
	RMP440_GYRO_ASSERV *gyro;
	gyro = &rmp440DataStrId->gyroAsserv;
	*yawr_command = gyro->prev_command;

	if (gyro->first)
	{
		*yawr_command = yawr_reference;
		gyro->first = 0;
		gyro->enabled = 1;
		gyro->jump_t = 0.;
	} else
	{
		double dt = t-gyro->prev_t;

		if (yawr_reference == 0.0)
		{
			if (vel_reference != 0.0)
			{ // straight -> directly asserv on angle
				if (gyro->straight)
				{
					if (yawr_measure == gyro->prev_measure) return;

					// PI controller on angle commanded in speed
					// I is needed to remove static error
					// D is not needed because only used to go straight (small errors)
					double error_s = (gyro->straight_angle-yaw_measure);
					if (error_s > PI) error_s -= 2*PI;
					if (error_s < -PI) error_s += 2*PI;
					gyro->integral_s += error_s * (t-gyro->prev_t);
					*yawr_command =
						RMP440_KP_GYRO_THETA*dt * error_s +
						RMP440_KI_GYRO_THETA*dt * gyro->integral_s;
				} else
				{
					*yawr_command = 0.0;
					gyro->straight = 1;
					gyro->straight_angle = yaw_measure;
					gyro->integral_s = 0.0;
				}
			} else
			{ // do nothing, command (0,0) is correctly respected
				*yawr_command = 0.0;
				gyro->straight = 0;
			}
		} else
		{
			const double ref_ratio = 1.5;
			const double noise_th = 0.1;

			// if changing sign or increasing from around 0
			// then directly apply a priori command
			if ((yawr_reference*gyro->prev_reference < 0) ||
			    ((fabs(gyro->prev_reference) < noise_th) && (fabs(yawr_reference) >= noise_th*ref_ratio)))
			{
				*yawr_command = yawr_reference;
				gyro->jump_t = t;
			}
			// if changing more than 50%
			// then directly apply extrapolated command
			if ((fabs(gyro->prev_reference) >= noise_th) &&
			    (yawr_reference*gyro->prev_reference > 0) &&
			    ((yawr_reference/gyro->prev_reference > ref_ratio) || (gyro->prev_reference/yawr_reference > ref_ratio)))
			{
				*yawr_command = gyro->prev_command * yawr_reference / gyro->prev_reference;
				gyro->jump_t = t;
			}
			// otherwise if delay from last jump has elapsed
			// P controller on speed commanded in speed
			if (t-gyro->jump_t >= RMP440_DELAY_GYRO_OMEGA)
			{
				if (yawr_measure == gyro->prev_measure) return;

				*yawr_command =
					gyro->prev_command +
					RMP440_KP_GYRO_OMEGA*dt * (yawr_reference-yawr_measure);
			} else
			// if we are in the lock period and we didn't just reenter it,
			// then we have to follow the reference changes
			if (t != gyro->jump_t)
			{
				if (fabs(gyro->prev_reference) < noise_th || fabs(yawr_reference - gyro->prev_reference) < 1e-10)
					*yawr_command = gyro->prev_command;
				else
					*yawr_command = gyro->prev_command * yawr_reference / gyro->prev_reference;
			}
			gyro->straight = 0;
		}
	}

//*yawr_command = yawr_reference; // TEST remove asserv

	gyro->prev_reference = yawr_reference;
	gyro->prev_measure = yawr_measure;
	gyro->prev_command = *yawr_command;
	gyro->prev_t = t;
}


/*------------------------------------------------------------------------
 *
 * asserv  --  2nd codel of permanent activity (called AFTER other activities)
 *
 * Description: Computes reference speeds and apply them
 *
 * Reports:      OK
 *              S_rmp440_EMERGENCY_STOP
 *
 * Returns:    OK or ERROR
 */

STATUS
asserv(int *report)
{
	GENPOS_CART_REF *refId = &SDI_F->ref;
	GENPOS_CART_STATE *robot = &SDI_F->robot;
	double vRef, wRef;

	if (rmp == NULL)
		return OK;		/* not initialized yet */

	*report = OK;
	switch (SDI_F->status.rs_mode) {
	case RMP440_MOTORS_OFF:
		/* No motion possible */
		return OK;

	case RMP440_EMERGENCY:
	case RMP440_IDLE:
		vRef = 0.0;
		wRef = 0.0;
		refId->x = robot->xRef;
		refId->y = robot->yRef;
		refId->theta = robot->theta;
		break;

	case RMP440_MANUAL:
		if (joystickPoll(joy) < 0) {
			*report = S_rmp440_JOYSTICK_ERROR;
			break;
		}
		refId->x = robot->xRef;
		refId->y = robot->yRef;
		refId->theta = robot->theta;
		getJoystickSpeeds(&vRef, &wRef, &refId->linAccelMax, &refId->angAccelMax);
		refId->v = vRef;
		refId->w = wRef;
		break;

	case RMP440_TRACK:
		track(report, &vRef, &wRef);
		break;

	default:
		*report = S_rmp440_BAD_REF;
		return ERROR;
	}
	if (*report != OK) {
		/* In case an error occured,
		   stop the robot and the tracking */
		SDI_F->status.rs_mode = RMP440_IDLE;
		SDI_F->trackStr.trackMode = GENPOS_NO_TRACKING;
		vRef = 0;
		wRef = 0;
	}

	SDI_F->vReference = vRef;
	SDI_F->wReference = wRef;

	H2TIMESPEC tv;
	h2GetTimeSpec(&tv);
	double t = tv.tv_sec + tv.tv_nsec*1e-9;

	//if (SDI_F->status.rs_mode == RMP440_TRACK)
	bound_accels(t, &vRef, &wRef);

#ifdef GYRO
	if (SDI_F->gyro.gyroOn)
		control_yaw(report, t, vRef, wRef, SDI_F->gyro.gyroOmega, SDI_F->gyro.gyroTheta, &wRef);
	//(SDI_F->status.rs_data[0].yaw_rate + SDI_F->status.rs_data[1].yaw_rate)/2.0; // do not use internal gyros, they are wrong
#endif

	rmp440VelocitySet(vRef, wRef, report);
	updatePomPosters();

	return OK;
}

/*------------------------------------------------------------------------
 * Init
 *
 * Description:
 *
 * Reports:      OK
 */

/* rmp440InitMain  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp440InitStart(int *report)
{

	*report = OK;
#ifdef WITH_FELIB
	/* connect emergency stop */
	fe = fe_init();
	if (fe == NULL) {
		*report = S_rmp440_EMERGENCY_ERROR;
		return ETHER;
	}
#endif
	/* Disconnect first if already connected */
	if (data != NULL && rmp != NULL) {
		if (data->operational_state == 4)
			rmp440SetOperationalMode(rmp, RMP_STANDBY_REQUEST);
		free(data);
		rmp440End(rmp);
	}
	data = rmp440FeedbackInit(NULL);
	if (data == NULL) {
		*report = S_rmp440_RMPLIB_ERROR;
		return ETHER;
	}

	rmp = rmp440Init(RMP440_INTERFACE_ETHERNET, "192.168.0.40:8080");
	if (rmp == NULL) {
		*report = S_rmp440_RMPLIB_ERROR;
		return ETHER;
	}
	rmp440SetOperationalMode(rmp, RMP_TRACTOR_REQUEST);
	return EXEC;
}

ACTIVITY_EVENT
rmp440InitMain(int *report)
{
	RMP440_STATUS_STR *status = &SDI_F->status;

	rmp440ReceiveAndDecode(rmp, data);

	/* Check motors status */
	if (data->operational_state == 4) {
		printf("Motors ON\n");
		status->rs_mode = RMP440_IDLE;
		return ETHER;
	}
	rmp440CmdNone(rmp);
	return EXEC;
}

/*------------------------------------------------------------------------
 * JoystickOn
 *
 * Description:
 *
 * Reports:      OK
 *              S_rmp440_BAD_REF
 *              S_rmp440_RMP_LIB_ERROR
 */

/* rmp440JoystickOnStart  -  codel START of JoystickOn
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp440JoystickOnStart(int *report)
{

	joy = joystickInit(RMP440_JOYSTICK_DEVICE_PATH);
	if (joy == NULL) {
		fprintf(stderr, "joystickInit(%s) failed: %s\n",
			RMP440_JOYSTICK_DEVICE_PATH, strerror(errno));
		*report = S_rmp440_JOYSTICK_ERROR;
		return ETHER;
	}
	joystickPoll(joy);
	memset(joy->buttons, 0, sizeof(joy->buttons));
	SDI_F->status.rs_mode = RMP440_MANUAL;
	return EXEC;
}

/* rmp440JoystickOnMain  -  codel EXEC of JoystickOn
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp440JoystickOnMain(int *report)
{
	RMP440_STATUS_STR *status = &rmp440DataStrId->status;

	/* Check if mode changed */
	switch (status->rs_mode) {
	case RMP440_POWER_CORD:
		*report = S_rmp440_POWER_CORD_CONNECTED;
		return END;
	case RMP440_EMERGENCY:
		*report = S_rmp440_EMERGENCY_STOP;
		return END;
	case RMP440_MOTORS_OFF:
		*report = S_rmp440_MOTORS_OFF;
		return END;
	default:
		break;
	}
	/* Check for user abort (button 4) */
	if (joy->buttons[3]) {
		printf("Stop joystick\n");
		return rmp440JoystickOnInter(report);
	}
	return EXEC;
}

/* rmp440JoystickOnInter  -  codel INTER of JoystickOn
   Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp440JoystickOnInter(int *report)
{
	GENPOS_CART_REF *refId;

	joystickPoll(joy);

	refId = &rmp440DataStrId->ref;
	refId->linAccelMax = RMP_DEFAULT_MAXIMUM_ACCEL;
	refId->angAccelMax = RMP_DEFAULT_MAXIMUM_YAW_RATE;
	SDI_F->status.rs_mode = RMP440_IDLE;
	joystickEnd(joy);
	joy = NULL;
	return ETHER;
}

/*------------------------------------------------------------------------
 * SetPom
 *
 * Description:
 *
 * Reports:      OK
 *              S_rmp440_POSTER_NOT_FOUND
 */

/* setPom  -  codel EXEC of SetPom
   Returns:  EXEC END ETHER FAIL ZOMBIE */

ACTIVITY_EVENT
setPom(GENPOS_POSTER_NAME *posterName, int *report)
{
	POSTER_ID id;

	if (posterFind(posterName->name, &id) != OK) {
		logMsg("posterFind(%s) failed\n", posterName->name);
		*report = S_rmp440_POSTER_NOT_FOUND;
	} else
		pomPosterId = id;
	return ETHER;
}

/*----------------------------------------------------------------------*/
/*
 * Compute speed
 *
 * sets robot.v and robot.w in the SDI from motion in the motor structs.
 */


static STATUS
rmp440VelocityGet(void)
{
	double leftv, rightv;
	GENPOS_CART_STATE *robot = &SDI_F->robot;
	RMP440_KINEMATICS_STR *kinematics = &SDI_F->kinematics;

	/* Mean angular speed of front and back wheels */
	leftv = (data->left_front_vel + data->left_rear_vel) / 2.0;
	rightv = (data->right_front_vel + data->right_rear_vel) / 2.0;
	/* printf("%.2lf %.2lf\n", leftv, rightv); */

	/* Angular speed of the robot */
#ifdef USE_INTERNAL_GYRO

	robot->w = data->pse_yaw_rate;
#else
	robot->w = (leftv - rightv) / kinematics->axisWidth;
#endif
	/* linear speeds */
	robot->v = data->linear_vel;

	return OK;
}

/*----------------------------------------------------------------------*/
/*
 * Send given linear and angular speeds to the hardware
 *
 */
static STATUS
rmp440VelocitySet(double v, double w, int *report)
{

	/* Last safety check, in case uninitialized values are used */
	if (fabs(v) > data->fram_vel_limit) {
		fprintf(stderr, "WARNING: v > VMAX: %lf\n", v);
		v = v > 0.0 ? 1.0 : -1.0;
	}
	/* Normalize */
	v = v/data->fram_vel_limit;

	if (fabs(w) > data->fram_yaw_rate_limit) {
		fprintf(stderr, "WARNING: w > WMAX: %lf\n", w);
		w = w > 0.0 ? 1.0 : -1.0;
	}
	/* Normalize */
	w = w/data->fram_yaw_rate_limit;

	/* Send commands to motors if ON */
	if (data->operational_state == 4)
		rmp440Motion(rmp, v, w);
	else
		rmp440CmdNone(rmp);

	/* Save the set velocities */
	rmp440DataStrId->vCommand = v;
	rmp440DataStrId->wCommand = w;

	return OK;
}

/*----------------------------------------------------------------------*/
/*
 * Initialize POM
 */
static STATUS
initPomPosters(void)
{
	POM_ME_POS *mePos = &rmp440DataStrId->mePos;

	mePos->kind = POM_ME_ABSOLUTE;
	mePos->date1 = -1;
	mePos->confidence = 0.2;	// XXX used?
	mePos->main.var.voo     = RMP440_VAR_TT;
	mePos->main.var.vpp     = RMP440_VAR_TT;
	mePos->main.var.vrr     = RMP440_VAR_TT;
	mePos->main.var.vxx     = RMP440_VAR_XX;
	mePos->main.var.vyy     = RMP440_VAR_YY;
	mePos->main.var.vzz     = RMP440_VAR_XX;

	return OK;
}

/*----------------------------------------------------------------------*/

/*
 * Update POM poster
 */
static STATUS
updatePomPosters(void)
{
	POM_ME_POS *mePos = &SDI_F->mePos;
	GENPOS_CART_STATE *configId;
	RMP440_CART_3DSTATE *config3dId;
	POM_POS pomPos;

	if (pomPosterId != NULL) {
		if(rmp440POM_POSPosterRead(pomPosterId, &pomPos) == OK) {
			mePos->date1 = pomPos.date;
		} else {
			logMsg("rmp440POM_POSPosterRead() failed\n");
			pomPosterId = NULL;
			mePos->date1 = -1;
		}
	} else
		mePos->date1 = -1;

	configId = &rmp440DataStrId->robot;
	mePos->main.euler.yaw = configId->theta;
	mePos->main.euler.x   = configId->xRob;
	mePos->main.euler.y   = configId->yRob;

	config3dId = &rmp440DataStrId->robot3d;
	mePos->main.euler.pitch = config3dId->pitch;
	mePos->main.euler.roll = config3dId->roll;
	mePos->main.euler.z = config3dId->zRob;

	/* speeds are mostly zero, just update the linear
	   and angular components */
	mePos->VLocal.euler.x   = configId->v;
	mePos->VLocal.euler.yaw = configId->w;

	/* all variances and co-variances remain zero */

	return OK;
}

/*----------------------------------------------------------------------*/

/*
 * Convert joystick values to reference speed
 *
 * A square function is used to get more precision at low speeds.
 *
 */
static void
getJoystickSpeeds(double *v, double *w, double *avmax, double *awmax)
{
	double vmaxw = -2, vmaxv = -2, againw = -2, againv = -2;
	if (joy->buttons[0]) { vmaxw = 0.7; vmaxv = 0.5; againw = 0.6; againv = 0.6; }
	if (joy->buttons[1]) { vmaxw = 0.9; vmaxv = 1.3; againw = 0.8; againv = 0.8; }
	if (joy->buttons[2]) { vmaxw = 1.1; vmaxv = 2.0; againw = 1.0; againv = 1.0; }
	if (vmaxv > -1) {
		*v = -SIGN(joy->axes[1]) * joy->axes[1] * joy->axes[1]
		  * RMP440_JOYSTICK_VGAIN * vmaxv;
		*w = -SIGN(joy->axes[0]) * joy->axes[0] * joy->axes[0]
		  * RMP440_JOYSTICK_WGAIN * vmaxw;
		if (avmax) *avmax = data->fram_accel_limit * againv;
		if (awmax) *awmax = data->fram_yaw_accel_limit * againw;
	}
	else {
		*v = 0;
		*w = 0;
	}
}
/*------------------------------------------------------------------------
 * Gyro
 *
 * Description: Setup gyroscope
 *
 * Reports:      OK
 *              S_rmp440_GYRO_ERROR
 */

/* rmp440GyroExec  -  codel EXEC of Gyro
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp440GyroExec(RMP440_GYRO_PARAMS *gyroParams, int *report)
{
	RMP440_GYRO *gyro = &SDI_F->gyro;

	if (gyroParams->mode == gyro->currentMode)
		return ETHER;

	if (gyroParams->mode == RMP440_GYRO_OFF) {
		gyro->currentMode = RMP440_GYRO_OFF;
		return ETHER;
	}

#ifdef GYRO
	if (gyroId == NULL) {
		gyroId = gyroInit(gyroParams->type, gyroParams->port,
			    gyroParams->latitude, gyroParams->woffset);
		if (gyroId == NULL) {
			*report = S_rmp440_GYRO_ERROR;
			gyro->params.mode = RMP440_GYRO_OFF;
			gyro->currentMode = RMP440_GYRO_OFF;
			return ETHER;
		}
	}

	/* read gyro once */
	if (!(gyroReadAngle(gyroId, &gyro->gyroTheta))) {
		*report = S_rmp440_GYRO_ERROR;
		gyro->params.mode = RMP440_GYRO_OFF;
		gyro->currentMode = RMP440_GYRO_OFF;
		gyroEnd(gyroId);
		gyroId = NULL;
		return ETHER;
	}

	/* reset gyro offset to match odo */
	gyro->gyroToRobotOffset = SDI_F->robot.theta - gyro->gyroTheta;

	gyro->currentMode = gyro->params.mode;

#else
	printf("rmp440GyroExec: rmp440 module compiled without gyro support. "
	    "Use -DGYRO\n");
	gyro->params.mode = RMP440_GYRO_OFF;
	gyro->currentMode = RMP440_GYRO_OFF;
	*report = S_rmp440_GYRO_ERROR;
#endif

	return ETHER;
}


/*------------------------------------------------------------------------
 * Log
 *
 * Description:
 *
 * Reports:      OK
 *              S_rflex_CANNOT_OPEN_FILE
 *              S_rflex_CANNOT_WRITE_FILE
 */

static FILE *rmp440LogFile;

/* rmp440LogStart  -  codel START of Log
	Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp440LogStart(GEN_STRING128 *fileName, int *report)
{
	if (!(rmp440LogFile = fopen(fileName->name, "a"))) {
		printf ("rmp440LogStart: cannot open odometer file %s\n", fileName->name);
		*report = S_rmp440_CANNOT_OPEN_FILE;
		return ETHER;
	}

	if (!fprintf(rmp440LogFile, "#date\tdate1\tdate2\tmain/yaw\tmain/pitch\tmain/roll\tmain/x\tmain/y\tmain/z\tvlocal/yaw\tvlocal/pitch\tvlocal/roll\tvlocal/x\tvlocal/y\tvlocal/z\n")) {
		printf ("rmp440LogStart: cannot log in %s\n", fileName->name);
		*report = S_rmp440_CANNOT_WRITE_FILE;
		rmp440LogInter(fileName, report);
		return ETHER;
	}

	printf ("rmp440LogStart: log file %s found\n", fileName->name);
	return EXEC;
}

/* rmp440LogExec  -  codel EXEC of Log
	Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp440LogExec(GEN_STRING128 *fileName, int *report)
{
	H2TIMESPEC tv;

	/* get date */
	h2GetTimeSpec(&tv);

	if (!fprintf(rmp440LogFile, "%lf\t%d\t%d\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\t%2.6g\n",
		tv.tv_sec + tv.tv_nsec*1.0E-9, SDI_F->mePos.date1, SDI_F->mePos.date2,
		SDI_F->mePos.main.euler.yaw, SDI_F->mePos.main.euler.pitch, SDI_F->mePos.main.euler.roll,
		SDI_F->mePos.main.euler.x, SDI_F->mePos.main.euler.y, SDI_F->mePos.main.euler.z,
		SDI_F->mePos.VLocal.euler.yaw, SDI_F->mePos.VLocal.euler.pitch, SDI_F->mePos.VLocal.euler.roll,
		SDI_F->mePos.VLocal.euler.x, SDI_F->mePos.VLocal.euler.y, SDI_F->mePos.VLocal.euler.z
		)) {
		printf ("rmp440LogExec: cannot log in %s\n", fileName->name);
		*report = S_rmp440_CANNOT_WRITE_FILE;
		rmp440LogInter(fileName, report);
		return ETHER;
	}

	return EXEC;
}


/* rmp440LogInter  -  codel INTER of Log
	Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp440LogInter(GEN_STRING128 *fileName, int *report)
{
	if (rmp440LogFile) {
		fclose(rmp440LogFile);
		rmp440LogFile = NULL;
	}
	return ETHER;
}

/*------------------------------------------------------------------------
 * LogInternalData
 *
 * Description: Log all the raw data provided by robot
 *
 * Reports:     OK
 *              S_rmp440_CANNOT_OPEN_FILE
 *              S_rmp440_CANNOT_WRITE_FILE
 */

static FILE *rmp440LogInternalDataFile;

/* rmp440LogStart  -  codel START of Log
	Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp440LogInternalDataStart(GEN_STRING128 *fileName, int *report)
{
  if (!(rmp440LogInternalDataFile = fopen(fileName->name, "a"))) {
    printf ("rmp440LogInternalDataStart: cannot open internal data file %s\n", fileName->name);
    *report = S_rmp440_CANNOT_OPEN_FILE;
    return ETHER;
  }

  if (!fprintf(rmp440LogInternalDataFile,
	       "#date\tdate1\tdate2"
	       "\tgyroAngle\tgyroRate"
#if (GYRO==gyroKVH1000Task)
	       "\tgyroTemperature"
#endif
	       "\tsent_velocity_reference\tsent_velocity_command"
	       "\tsent_turn_reference\tsent_turn_command"
	       "\tstraight\tstraight_angle"
	       "\tintegrated_yaw1\tyaw_rate1"
	       "\tpitch_angle1[broken]\tpitch_rate1"
	       "\troll_angle1\troll_rate1"
	       "\tlw_velocity1\trw_velocity1"
	       "\tintegrated_left_wheel1\tintegrated_right_wheel1"
	       "\tintegrated_fore_aft1\tleft_torque1\tright_torque1"
	       "\tservo_frames1\toperational_mode1"
	       "\tcontroller_gain_schedule1\tui_voltage1"
	       "\tpowerbase_voltage1\tbattery_charge1"
	       "\tvelocity_command1[broken]\tturn_command1[broken]"
	       "\tintegrated_yaw2\tyaw_rate2"
	       "\tpitch_angle2[broken]\tpitch_rate2"
	       "\troll_angle2\troll_rate2"
	       "\tlw_velocity2\trw_velocity2"
	       "\tintegrated_left_wheel2\tintegrated_right_wheel2"
	       "\tintegrated_fore_aft2\tleft_torque2\tright_torque2"
	       "\tservo_frames2\toperational_mode2"
	       "\tcontroller_gain_schedule2\tui_voltage2"
	       "\tpowerbase_voltage2\tbattery_charge2"
	       "\tvelocity_command2[broken]\tturn_command2[broken]\n")) {
    printf ("rmp440LogStart: cannot log in %s\n", fileName->name);
    *report = S_rmp440_CANNOT_WRITE_FILE;
    rmp440LogInternalDataInter(fileName, report);
    return ETHER;
  }
  printf ("rmp440LogInternalDataStart: log file %s found\n", fileName->name);
  return EXEC;
}

/* rmp440LogInternalDataExec  -  codel EXEC of Log
	Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp440LogInternalDataExec(GEN_STRING128 *fileName, int *report)
{
  H2TIMESPEC tv;

  /* get date */
  h2GetTimeSpec(&tv);
#if 0
  if (!fprintf(rmp440LogInternalDataFile, "%lf\t%d\t%d"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g\t%d"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%d\t%d"
	       "\t%d\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g\t%.6g"
	       "\t%d\t%d"
	       "\t%d\t%.6g"
	       "\t%.6g\t%.6g"
	       "\t%.6g\t%.6g"
	       "\n",
				/* timestamps */
	       tv.tv_sec + tv.tv_nsec*1.0E-9,
	       SDI_F->mePos.date1, SDI_F->mePos.date2,
				/* the KVH gyro */
	       SDI_F->gyro.gyroTheta,
	       SDI_F->gyro.gyroOmega,
	       SDI_F->vReference, SDI_F->vCommand, // need to check if latest value is logged
	       SDI_F->wReference, SDI_F->wCommand,
	       SDI_F->gyroAsserv.straight, SDI_F->gyroAsserv.straight_angle,
				/* the robot first axle */
	       SDI_F->status.rs_data[0].integrated_yaw,
	       SDI_F->status.rs_data[0].yaw_rate,
	       SDI_F->status.rs_data[0].pitch_angle,
	       SDI_F->status.rs_data[0].pitch_rate,
	       SDI_F->status.rs_data[0].roll_angle,
	       SDI_F->status.rs_data[0].roll_rate,
	       SDI_F->status.rs_data[0].rw_velocity, // invert left/right
	       SDI_F->status.rs_data[0].lw_velocity,
	       SDI_F->status.rs_data[0].integrated_right_wheel, // invert left/right
	       SDI_F->status.rs_data[0].integrated_left_wheel,
	       SDI_F->status.rs_data[0].integrated_fore_aft,
	       SDI_F->status.rs_data[0].right_torque, // invert left/right
	       SDI_F->status.rs_data[0].left_torque,
	       SDI_F->status.rs_data[0].servo_frames,
	       SDI_F->status.rs_data[0].operational_mode,
	       SDI_F->status.rs_data[0].controller_gain_schedule,
	       SDI_F->status.rs_data[0].ui_voltage,
	       SDI_F->status.rs_data[0].powerbase_voltage,
	       SDI_F->status.rs_data[0].battery_charge,
	       SDI_F->status.rs_data[0].velocity_command,
	       SDI_F->status.rs_data[0].turn_command,
				/* the robot second axle */
	       SDI_F->status.rs_data[1].integrated_yaw,
	       SDI_F->status.rs_data[1].yaw_rate,
	       SDI_F->status.rs_data[1].pitch_angle,
	       SDI_F->status.rs_data[1].pitch_rate,
	       SDI_F->status.rs_data[1].roll_angle,
	       SDI_F->status.rs_data[1].roll_rate,
	       SDI_F->status.rs_data[1].rw_velocity, // invert left/right
	       SDI_F->status.rs_data[1].lw_velocity,
	       SDI_F->status.rs_data[1].integrated_right_wheel, // invert left/right
	       SDI_F->status.rs_data[1].integrated_left_wheel,
	       SDI_F->status.rs_data[1].integrated_fore_aft,
	       SDI_F->status.rs_data[1].right_torque, // invert left/right
	       SDI_F->status.rs_data[1].left_torque,
	       SDI_F->status.rs_data[1].servo_frames,
	       SDI_F->status.rs_data[1].operational_mode,
	       SDI_F->status.rs_data[1].controller_gain_schedule,
	       SDI_F->status.rs_data[1].ui_voltage,
	       SDI_F->status.rs_data[1].powerbase_voltage,
	       SDI_F->status.rs_data[1].battery_charge,
	       SDI_F->status.rs_data[1].velocity_command,
	       SDI_F->status.rs_data[1].turn_command))
    {
      printf ("rmp440LogInternalDataExec: cannot log in %s\n",
	      fileName->name);
      *report = S_rmp440_CANNOT_WRITE_FILE;
      rmp440LogInter(fileName, report);
      return ETHER;
    }
#endif
  return EXEC;
}

/* rmp440LogInternalDataInter  -  codel INTER of LogInternalData
	Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rmp440LogInternalDataInter(GEN_STRING128 *fileName, int *report)
{
  if (rmp440LogInternalDataFile) {
    fclose(rmp440LogInternalDataFile);
    rmp440LogInternalDataFile = NULL;
  }
  return ETHER;
}



#ifdef WITH_FELIB
/*------------------------------------------------------------------------
 * ResetEmergencyStop
 *
 * Description:
 *
 * Reports:      OK
 *              S_rmp440_EMERGENCY_STOP
 */

/* resMain  -  codel EXEC of ResetEmergencyStop
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
resMain(int *report)
{
	RMP440_STATUS_STR *status = &SDI_F->status;
	int i;

	if ((fe_pins(fe) & (FE_PAUSE|FE_STOP)) != 0) {
		*report = S_rmp440_EMERGENCY_STOP;
		return END;
	}
	if ((fe_pins(fe) & FE_POWER_CORD) != 0)
		status->rs_mode = RMP440_POWER_CORD;
	else
		status->rs_mode = RMP440_IDLE;

	return ETHER;
}
#endif
